@@ -27,7 +27,7 @@ import codecs
 import platform
 import importlib
 from gzip import GzipFile
-from base64 import b64decode, b64encode
+from base64 import b64encode
 from decimal import Context, Decimal, Inexact
 from hashlib import sha1
 from collections import namedtuple
@@ -107,6 +107,9 @@ def quiet_raise(exc):
     def zero_byte_sequence(n):
         return bytes(n)
 
+    def int_to_byte(i):
+        return bytes([i])
+
     import struct
 
     struct_pack = struct.pack
@@ -159,6 +162,8 @@ else:
             raise OverflowError('int too big to convert')
         return hbytes(result)
 
+    int_to_byte = chr
+
     def bytes_from_list(ls):
         return hbytes(bytearray(ls))
 
@@ -290,16 +295,8 @@ else:
         getfullargspec = silence_warnings(getfullargspec)
 
 
-def get_type_hints(thing):
-    """Try to return any type hints for ``thing``."""
-    try:
-        import typing
-        return typing.get_type_hints(thing)
-    except TypeError:
-        # `thing` is not a module, class, method, or function
-        return {}
-    except (ImportError, AttributeError):  # pragma: no cover
-        # This is a fallback for Python <3.6, where get_type_hints may fail
+if sys.version_info[:2] < (3, 6):
+    def get_type_hints(thing):
         try:
             spec = getfullargspec(thing)
             return {
@@ -308,6 +305,13 @@ def get_type_hints(thing):
             }
         except TypeError:
             return {}
+else:
+    def get_type_hints(thing):
+        try:
+            import typing
+            return typing.get_type_hints(thing)
+        except TypeError:
+            return {}
 
 
 importlib_invalidate_caches = getattr(
@@ -502,7 +506,26 @@ else:
 
 
 if PY2:
+    # Under Python 2, math.floor and math.ceil return floats, which cannot
+    # represent large integers - eg `float(2**53) == float(2**53 + 1)`.
+    # We therefore implement them entirely in (long) integer operations.
     def floor(x):
-        return int(math.floor(x))
+        if int(x) != x and x < 0:
+            return int(x) - 1
+        return int(x)
+
+    def ceil(x):
+        if int(x) != x and x > 0:
+            return int(x) + 1
+        return int(x)
 else:
     floor = math.floor
+    ceil = math.ceil
+
+
+if PY2:
+    def b64decode(s):
+        from base64 import b64decode as base
+        return hbytes(base(s))
+else:
+    from base64 import b64decode
