@@ -21,18 +21,20 @@ import math
 import datetime as dt
 import operator
 from decimal import Decimal, InvalidOperation
+from inspect import isclass
 from numbers import Rational
 from fractions import Fraction
 
-from hypothesis.errors import InvalidArgument
+from hypothesis.errors import InvalidArgument, ResolutionFailed
 from hypothesis.control import assume
+from hypothesis._settings import note_deprecation
 from hypothesis.searchstrategy import SearchStrategy
-from hypothesis.internal.compat import hrange, text_type, integer_types, \
-    getfullargspec, implements_iterator
+from hypothesis.internal.compat import ceil, floor, hrange, text_type, \
+    integer_types, get_type_hints, getfullargspec, implements_iterator
 from hypothesis.internal.floats import is_negative, float_to_int, \
     int_to_float, count_between_floats
-from hypothesis.utils.conventions import not_set
-from hypothesis.internal.reflection import proxies
+from hypothesis.utils.conventions import infer, not_set
+from hypothesis.internal.reflection import proxies, required_args
 from hypothesis.searchstrategy.reprwrapper import ReprWrapperStrategy
 
 __all__ = [
@@ -42,7 +44,7 @@ __all__ = [
     'choices', 'streaming',
     'booleans', 'integers', 'floats', 'complex_numbers', 'fractions',
     'decimals',
-    'characters', 'text', 'binary', 'uuids',
+    'characters', 'text', 'from_regex', 'binary', 'uuids',
     'tuples', 'lists', 'sets', 'frozensets', 'iterables',
     'dictionaries', 'fixed_dictionaries',
     'sampled_from', 'permutations',
@@ -52,6 +54,7 @@ __all__ = [
     'recursive', 'composite',
     'shared', 'runner', 'data',
     'deferred',
+    'from_type', 'register_type_strategy',
 ]
 
 _strategies = set()
@@ -101,6 +104,7 @@ def cacheable(fn):
             result = fn(*args, **kwargs)
             cache[cache_key] = result
             return result
+    cached_strategy.__clear_cache = cache.clear
     return cached_strategy
 
 
@@ -201,17 +205,13 @@ def integers(min_value=None, max_value=None):
     from hypothesis.searchstrategy.numbers import IntegersFromStrategy, \
         BoundedIntStrategy, WideRangeIntStrategy
 
-    min_int_value = None
-    if min_value is not None:
-        min_int_value = int(min_value)
-        if min_int_value != min_value and min_value > 0:
-            min_int_value += 1
+    min_int_value = None if min_value is None else ceil(min_value)
+    max_int_value = None if max_value is None else floor(max_value)
 
-    max_int_value = None
-    if max_value is not None:
-        max_int_value = int(max_value)
-        if max_int_value != max_value and max_value < 0:
-            max_int_value -= 1
+    if min_int_value is not None and max_int_value is not None and \
+            min_int_value > max_int_value:
+        raise InvalidArgument('No integers between min_value=%r and '
+                              'max_value=%r' % (min_value, max_value))
 
     if min_int_value is None:
         if max_int_value is None:
@@ -605,7 +605,6 @@ def dictionaries(
     ).map(dict_class)
 
 
-@cacheable
 @defines_strategy
 def streaming(elements):
     """Generates an infinite stream of values where each value is drawn from
@@ -614,8 +613,17 @@ def streaming(elements):
     The result is iterable (the iterator will never terminate) and
     indexable.
 
+    .. deprecated:: 3.15.0
+        Use :func:`data() <hypothesis.strategies.data>` instead.
+
     """
     check_strategy(elements)
+
+    note_deprecation(
+        'streaming() has been deprecated. Use the data() strategy instead and '
+        'replace stream iteration with data.draw() calls.'
+    )
+
     from hypothesis.searchstrategy.streams import StreamStrategy
     return StreamStrategy(elements)
 
@@ -624,19 +632,30 @@ def streaming(elements):
 @defines_strategy
 def characters(whitelist_categories=None, blacklist_categories=None,
                blacklist_characters=None, min_codepoint=None,
-               max_codepoint=None):
+               max_codepoint=None, whitelist_characters=None):
     """Generates unicode text type (unicode on python 2, str on python 3)
     characters following specified filtering rules.
 
-    This strategy accepts lists of Unicode categories, characters of which
-    should (`whitelist_categories`) or should not (`blacklist_categories`)
-    be produced.
+    When no filtering rules are specifed, any character can be produced.
 
-    Also there could be applied limitation by minimal and maximal produced
-    code point of the characters.
+    If ``min_codepoint`` or ``max_codepoint`` is specifed, then only
+    characters having a codepoint in that range will be produced.
 
-    If you know what exactly characters you don't want to be produced,
-    pass them with `blacklist_characters` argument.
+    If ``whitelist_categories`` is specified, then only characters from those
+    Unicode categories will be produced. This is a further restriction,
+    characters must also satisfy `min_codepoint` and `max_codepoint`.
+
+    If ``blacklist_categories`` is specified, then any character from those
+    categories will not be produced. This is a further restriction,
+    characters that match both ``whitelist_categories`` and
+    ``blacklist_categories`` will not be produced.
+
+    If ``whitelist_characters`` is specified, then any additional characters
+    in that list will also be produced.
+
+    If ``blacklist_characters`` is specified, then any characters in that list
+    will be not be produced. Any overlap between ``whitelist_characters`` and
+    ``blacklist_characters`` will raise an exception.
 
     """
     if (
@@ -648,13 +667,40 @@ def characters(whitelist_categories=None, blacklist_categories=None,
                 min_codepoint, max_codepoint
             )
         )
+    if all((whitelist_characters is not None,
+            min_codepoint is None,
+            max_codepoint is None,
+            whitelist_categories is None,
+            blacklist_categories is None,
+            )):
+        raise InvalidArgument(
+            'Cannot have just whitelist_characters=%r alone, '
+            'it would have no effect. Perhaps you want sampled_from()' % (
+                whitelist_characters,
+            )
+        )
+    if (
+        whitelist_characters is not None and
+        blacklist_characters is not None and
+        set(blacklist_characters).intersection(set(whitelist_characters))
+    ):
+        raise InvalidArgument(
+            'Characters %r are present in both whitelist_characters=%r, and '
+            'blacklist_characters=%r' % (
+                set(blacklist_characters).intersection(
+                    set(whitelist_characters)
+                ),
+                whitelist_characters, blacklist_characters,
+            )
+        )
 
     from hypothesis.searchstrategy.strings import OneCharStringStrategy
     return OneCharStringStrategy(whitelist_categories=whitelist_categories,
                                  blacklist_categories=blacklist_categories,
                                  blacklist_characters=blacklist_characters,
                                  min_codepoint=min_codepoint,
-                                 max_codepoint=max_codepoint)
+                                 max_codepoint=max_codepoint,
+                                 whitelist_characters=whitelist_characters)
 
 
 @cacheable
@@ -693,6 +739,37 @@ def text(
     ))
 
 
+@cacheable
+@defines_strategy
+def from_regex(regex):
+    """Generates strings that contain a match for the given regex (i.e. ones
+    for which re.search will return a non-None result).
+
+    ``regex`` may be a pattern or :func:`compiled regex <python:re.compile>`.
+    Both byte-strings and unicode strings are supported, and will generate
+    examples of the same type.
+
+    You can use regex flags (such as :const:`python:re.IGNORECASE`,
+    :const:`python:re.DOTALL` or :const:`python:re.UNICODE`) to control
+    generation. Flags can be passed either in compiled regex or inside the
+    pattern with a ``(?iLmsux)`` group.
+
+    Some regular expressions are only partly supported - the underlying
+    strategy checks local matching and relies on filtering to resolve
+    context-dependent expressions.  Using too many of these constructs may
+    cause health-check errors as too many examples are filtered out. This
+    mainly includes (positive or negative) lookahead and lookbehind groups.
+
+    If you want the generated string to match the whole regex you should use
+    boundary markers. So e.g. ``r"\\A.\\Z"`` will return a single character
+    string, while ``"."`` will return any string, and ``r"\\A.$"`` will return
+    a single character optionally followed by a ``"\\n"``.
+
+    """
+    from hypothesis.searchstrategy.regex import regex_strategy
+    return regex_strategy(regex)
+
+
 @cacheable
 @defines_strategy
 def binary(
@@ -767,18 +844,161 @@ def random_module():
 @cacheable
 @defines_strategy
 def builds(target, *args, **kwargs):
-    """Generates values by drawing from args and kwargs and passing them to
-    target in the appropriate argument position.
+    """Generates values by drawing from ``args`` and ``kwargs`` and passing
+    them to ``target`` in the appropriate argument position.
 
-    e.g. builds(target, integers(), flag=booleans()) would draw an
-    integer i and a boolean b and call target(i, flag=b).
+    e.g. ``builds(target, integers(), flag=booleans())`` would draw an
+    integer ``i`` and a boolean ``b`` and call ``target(i, flag=b)``.
+
+    If ``target`` has type annotations, they will be used to infer a strategy
+    for required arguments that were not passed to builds.  You can also tell
+    builds to infer a strategy for an optional argument by passing the special
+    value :const:`hypothesis.infer` as a keyword argument to
+    builds, instead of a strategy for that argument to ``target``.
 
     """
+    if infer in args:
+        # Avoid an implementation nightmare juggling tuples and worse things
+        raise InvalidArgument('infer was passed as a positional argument to '
+                              'builds(), but is only allowed as a keyword arg')
+    hints = get_type_hints(target.__init__ if isclass(target) else target)
+    for kw in [k for k, v in kwargs.items() if v is infer]:
+        if kw not in hints:
+            raise InvalidArgument(
+                'passed %s=infer for %s, but %s has no type annotation'
+                % (kw, target.__name__, kw))
+        kwargs[kw] = from_type(hints[kw])
+    required = required_args(target, args, kwargs)
+    for ms in set(hints) & (required or set()):
+        kwargs[ms] = from_type(hints[ms])
     return tuples(tuples(*args), fixed_dictionaries(kwargs)).map(
         lambda value: target(*value[0], **value[1])
     )
 
 
+def delay_error(func):
+    """A decorator to make exceptions lazy but success immediate.
+
+    We want from_type to resolve to a strategy immediately if possible,
+    for a useful repr and interactive use, but delay errors until a
+    value would be drawn to localise them to a particular test.
+
+    """
+    @proxies(func)
+    def inner(*args, **kwargs):
+        try:
+            return func(*args, **kwargs)
+        except Exception as e:
+            error = e
+
+            def lazy_error():
+                raise error
+
+            return builds(lazy_error)
+    return inner
+
+
+@cacheable
+@delay_error
+def from_type(thing):
+    """Looks up the appropriate search strategy for the given type.
+
+    ``from_type`` is used internally to fill in missing arguments to
+    :func:`~hypothesis.strategies.builds` and can be used interactively
+    to explore what strategies are available or to debug type resolution.
+
+    You can use :func:`~hypothesis.strategies.register_type_strategy` to
+    handle your custom types, or to globally redefine certain strategies -
+    for example excluding NaN from floats, or use timezone-aware instead of
+    naive time and datetime strategies.
+
+    The resolution logic may be changed in a future version, but currently
+    tries these four options:
+
+    1. If ``thing`` is in the default lookup mapping or user-registered lookup,
+       return the corresponding strategy.  The default lookup covers all types
+       with Hypothesis strategies, including extras where possible.
+    2. If ``thing`` is from the :mod:`python:typing` module, return the
+       corresponding strategy (special logic).
+    3. If ``thing`` has one or more subtypes in the merged lookup, return
+       the union of the strategies for those types that are not subtypes of
+       other elements in the lookup.
+    4. Finally, if ``thing`` has type annotations for all required arguments,
+       it is resolved via :func:`~hypothesis.strategies.builds`.
+
+    """
+    from hypothesis.searchstrategy import types
+    if not isinstance(thing, type):
+        # Under Python 3.6, Unions are not instances of `type` - but we still
+        # want to resolve them!  This __origin__ check only passes if thing is
+        # a Union with parameters; if it doesn't we can't resolve it anyway.
+        try:
+            import typing
+            if getattr(thing, '__origin__', None) is typing.Union:
+                args = sorted(thing.__args__, key=types.type_sorting_key)
+                return one_of([from_type(t) for t in args])
+        except ImportError:  # pragma: no cover
+            pass
+        raise InvalidArgument('thing=%s must be a type' % (thing,))
+    # Now that we know `thing` is a type, the first step is to check for an
+    # explicitly registered strategy.  This is the best (and hopefully most
+    # common) way to resolve a type to a strategy.  Note that the value in the
+    # lookup may be a strategy or a function from type -> strategy; and we
+    # convert empty results into an explicit error.
+    if thing in types._global_type_lookup:
+        strategy = types._global_type_lookup[thing]
+        if not isinstance(strategy, SearchStrategy):
+            strategy = strategy(thing)
+        if strategy.is_empty:
+            raise ResolutionFailed(
+                'Error: %r resolved to an empty strategy' % (thing,))
+        return strategy
+    # If there's no explicitly registered strategy, maybe a subtype of thing
+    # is registered - if so, we can resolve it to the subclass strategy.
+    # We'll start by checking if thing is from from the typing module,
+    # because there are several special cases that don't play well with
+    # subclass and instance checks.
+    try:
+        import typing
+        if isinstance(thing, typing.TypingMeta):
+            return types.from_typing_type(thing)
+    except ImportError:  # pragma: no cover
+        pass
+    # If it's not from the typing module, we get all registered types that are
+    # a subclass of `thing` and are not themselves a subtype of any other such
+    # type.  For example, `Number -> integers() | floats()`, but bools() is
+    # not included because bool is a subclass of int as well as Number.
+    strategies = [
+        v if isinstance(v, SearchStrategy) else v(thing)
+        for k, v in types._global_type_lookup.items()
+        if issubclass(k, thing) and
+        sum(types.try_issubclass(k, T) for T in types._global_type_lookup) == 1
+    ]
+    empty = ', '.join(repr(s) for s in strategies if s.is_empty)
+    if empty:
+        raise ResolutionFailed(
+            'Could not resolve %s to a strategy; consider using '
+            'register_type_strategy' % empty)
+    elif strategies:
+        return one_of(strategies)
+    # If we don't have a strategy registered for this type or any subtype, we
+    # may be able to fall back on type annotations.
+    # Types created via typing.NamedTuple use a custom attribute instead -
+    # but we can still use builds(), if we work out the right kwargs.
+    if issubclass(thing, tuple) and hasattr(thing, '_fields') \
+            and hasattr(thing, '_field_types'):
+        kwargs = {k: from_type(thing._field_types[k]) for k in thing._fields}
+        return builds(thing, **kwargs)
+    # If the constructor has an annotation for every required argument,
+    # we can (and do) use builds() without supplying additional arguments.
+    required = required_args(thing)
+    if not required or required.issubset(get_type_hints(thing.__init__)):
+        return builds(thing)
+    # We have utterly failed, and might as well say so now.
+    raise ResolutionFailed('Could not resolve %r to a strategy; consider '
+                           'using register_type_strategy' % (thing,))
+
+
 @cacheable
 @defines_strategy
 def fractions(min_value=None, max_value=None, max_denominator=None):
@@ -1123,16 +1343,26 @@ def shared(base, key=None):
     return SharedStrategy(base, key)
 
 
-@cacheable
+@defines_strategy
 def choices():
     """Strategy that generates a function that behaves like random.choice.
 
     Will note choices made for reproducibility.
 
+    .. deprecated:: 3.15.0
+
+        Use :func:`data() <hypothesis.strategies.data>` with
+        :func:`sampled_from() <hypothesis.strategies.sampled_from>` instead.
+
     """
     from hypothesis.control import note, current_build_context
     from hypothesis.internal.conjecture.utils import choice, check_sample
 
+    note_deprecation(
+        'choices() has been deprecated. Use the data() strategy instead and '
+        'replace its usage with data.draw(sampled_from(elements))) calls.'
+    )
+
     class Chooser(object):
 
         def __init__(self, build_context, data):
@@ -1158,11 +1388,10 @@ def choices():
         def do_draw(self, data):
             return Chooser(current_build_context(), data)
 
-    return ReprWrapperStrategy(
-        shared(
-            ChoiceStrategy(),
-            key='hypothesis.strategies.chooser.choice_function'
-        ), 'choices()')
+    return shared(
+        ChoiceStrategy(),
+        key='hypothesis.strategies.chooser.choice_function'
+    )
 
 
 @cacheable
@@ -1271,6 +1500,35 @@ def data():
             ) % (name,))
     return DataStrategy()
 
+
+def register_type_strategy(custom_type, strategy):
+    """Add an entry to the global type-to-strategy lookup.
+
+    This lookup is used in :func:`~hypothesis.strategies.builds` and
+    :func:`@given <hypothesis.given>`.
+
+    :func:`~hypothesis.strategies.builds` will be used automatically for
+    classes with type annotations on ``__init__`` , so you only need to
+    register a strategy if one or more arguments need to be more tightly
+    defined than their type-based default, or if you want to supply a strategy
+    for an argument with a default value.
+
+    ``strategy`` may be a search strategy, or a function that takes a type and
+    returns a strategy (useful for generic types).
+
+    """
+    from hypothesis.searchstrategy import types
+    if not isinstance(custom_type, type):
+        raise InvalidArgument('custom_type=%r must be a type')
+    elif not (isinstance(strategy, SearchStrategy) or callable(strategy)):
+        raise InvalidArgument(
+            'strategy=%r must be a SearchStrategy, or a function that takes '
+            'a generic type and returns a specific SearchStrategy')
+    elif isinstance(strategy, SearchStrategy) and strategy.is_empty:
+        raise InvalidArgument('strategy=%r must not be empty')
+    types._global_type_lookup[custom_type] = strategy
+    from_type.__clear_cache()
+
 # Private API below here
 
 
@@ -1379,23 +1637,21 @@ def deferred(definition):
 
     Example usage:
 
-    .. doctest::
-        >>> import hypothesis.strategies as st
-        >>> x = st.deferred(lambda: st.booleans() | st.tuples(x, x))
-        >>> x.example()
-        (False, (False, True))
-        >>> x.example()
-        True
+    >>> import hypothesis.strategies as st
+    >>> x = st.deferred(lambda: st.booleans() | st.tuples(x, x))
+    >>> x.example()
+    (False, (False, True))
+    >>> x.example()
+    True
 
     Mutual recursion also works fine:
 
-    .. doctest::
-        >>> a = st.deferred(lambda: st.booleans() | b)
-        >>> b = st.deferred(lambda: st.tuples(a, a))
-        >>> a.example()
-        (((True, True), False), True)
-        >>> b.example()
-        (((( False, ((True, False), (True, True))), True), False), True)
+    >>> a = st.deferred(lambda: st.booleans() | b)
+    >>> b = st.deferred(lambda: st.tuples(a, a))
+    >>> a.example()
+    (((True, True), False), True)
+    >>> b.example()
+    (((( False, ((True, False), (True, True))), True), False), True)
 
     """
     from hypothesis.searchstrategy.deferred import DeferredStrategy
